@@ -1,5 +1,5 @@
 /* Copyright (c) 2010, 2015, Oracle and/or its affiliates.
-   Copyright (c) 2011, 2018, MariaDB
+   Copyright (c) 2011, 2020, MariaDB
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -14,7 +14,8 @@
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1335  USA */
 
-#include "sql_class.h"                       // THD and my_global.h
+#include "mariadb.h"
+#include "sql_class.h"                       // THD
 #include "keycaches.h"                       // get_key_cache
 #include "sql_base.h"                        // Open_table_context
 #include "lock.h"                            // MYSQL_OPEN_*
@@ -31,7 +32,7 @@
 #include "strfunc.h"
 #include "sql_admin.h"
 #include "sql_statistics.h"
-
+#include "wsrep_mysqld.h"
 /* Prepare, run and cleanup for mysql_recreate_table() */
 
 static bool admin_recreate_table(THD *thd, TABLE_LIST *table_list)
@@ -76,7 +77,7 @@ static int send_check_errmsg(THD *thd, TABLE_LIST* table,
 {
   Protocol *protocol= thd->protocol;
   protocol->prepare_for_resend();
-  protocol->store(table->alias, system_charset_info);
+  protocol->store(table->alias.str, table->alias.length, system_charset_info);
   protocol->store((char*) operator_name, system_charset_info);
   protocol->store(STRING_WITH_LEN("error"), system_charset_info);
   protocol->store(errmsg, system_charset_info);
@@ -90,10 +91,10 @@ static int send_check_errmsg(THD *thd, TABLE_LIST* table,
 static int prepare_for_repair(THD *thd, TABLE_LIST *table_list,
 			      HA_CHECK_OPT *check_opt)
 {
-  int error= 0;
+  int error= 0, create_error= 0;
   TABLE tmp_table, *table;
   TABLE_LIST *pos_in_locked_tables= 0;
-  TABLE_SHARE *share;
+  TABLE_SHARE *share= 0;
   bool has_mdl_lock= FALSE;
   char from[FN_REFLEN],tmp[FN_REFLEN+32];
   const char **ext;
@@ -122,7 +123,7 @@ static int prepare_for_repair(THD *thd, TABLE_LIST *table_list,
     */
 
     table_list->mdl_request.init(MDL_key::TABLE,
-                                 table_list->db, table_list->table_name,
+                                 table_list->db.str, table_list->table_name.str,
                                  MDL_EXCLUSIVE, MDL_TRANSACTION);
 
     if (lock_table_names(thd, table_list, table_list->next_global,
@@ -134,7 +135,8 @@ static int prepare_for_repair(THD *thd, TABLE_LIST *table_list,
     if (share == NULL)
       DBUG_RETURN(0);				// Can't open frm file
 
-    if (open_table_from_share(thd, share, "", 0, 0, 0, &tmp_table, FALSE))
+    if (open_table_from_share(thd, share, &empty_clex_str, 0, 0, 0,
+                              &tmp_table, FALSE))
     {
       tdc_release_share(share);
       DBUG_RETURN(0);                           // Out of memory
@@ -173,7 +175,7 @@ static int prepare_for_repair(THD *thd, TABLE_LIST *table_list,
   /*
     Check if this is a table type that stores index and data separately,
     like ISAM or MyISAM. We assume fixed order of engine file name
-    extentions array. First element of engine file name extentions array
+    extensions array. First element of engine file name extensions array
     is meta/index file extention. Second element - data file extention. 
   */
   ext= table->file->bas_ext();
@@ -188,7 +190,7 @@ static int prepare_for_repair(THD *thd, TABLE_LIST *table_list,
   if (!mysql_file_stat(key_file_misc, from, &stat_info, MYF(0)))
     goto end;				// Can't use USE_FRM flag
 
-  my_snprintf(tmp, sizeof(tmp), "%s-%lx_%lx",
+  my_snprintf(tmp, sizeof(tmp), "%s-%lx_%llx",
 	      from, current_pid, thd->thread_id);
 
   if (table_list->table)
@@ -205,6 +207,23 @@ static int prepare_for_repair(THD *thd, TABLE_LIST *table_list,
                               HA_EXTRA_NOT_USED, NULL);
     table_list->table= 0;
   }
+  else
+  {
+    /*
+      Table open failed, maybe because we run out of memory.
+      Close all open tables and relaese all MDL locks
+    */
+#if MYSQL_VERSION < 100500
+    tdc_remove_table(thd, TDC_RT_REMOVE_UNUSED,
+                     table->s->db.str, table->s->table_name.str,
+                     TRUE);
+#else
+    tdc_release_share(share);
+    share->tdc->flush(thd, true);
+    share= 0;
+#endif
+  }
+
   /*
     After this point we have an exclusive metadata lock on our table
     in both cases when table was successfully open in mysql_admin_table()
@@ -217,12 +236,9 @@ static int prepare_for_repair(THD *thd, TABLE_LIST *table_list,
 			     "Failed renaming data file");
     goto end;
   }
-  if (dd_recreate_table(thd, table_list->db, table_list->table_name))
-  {
-    error= send_check_errmsg(thd, table_list, "repair",
-			     "Failed generating table from .frm file");
-    goto end;
-  }
+  if (dd_recreate_table(thd, table_list->db.str, table_list->table_name.str))
+    create_error= send_check_errmsg(thd, table_list, "repair",
+                                    "Failed generating table from .frm file");
   /*
     'FALSE' for 'using_transactions' means don't postpone
     invalidation till the end of a transaction, but do it
@@ -235,6 +251,8 @@ static int prepare_for_repair(THD *thd, TABLE_LIST *table_list,
 			     "Failed restoring .MYD file");
     goto end;
   }
+  if (create_error)
+    goto end;
 
   if (thd->locked_tables_list.locked_tables())
   {
@@ -262,10 +280,11 @@ end:
   if (table == &tmp_table)
   {
     closefrm(table);
-    tdc_release_share(table->s);
+    if (share)
+      tdc_release_share(share);
   }
   /* In case of a temporary table there will be no metadata lock. */
-  if (error && has_mdl_lock)
+  if (unlikely(error) && has_mdl_lock)
     thd->release_transactional_locks();
 
   DBUG_RETURN(error);
@@ -304,7 +323,7 @@ static bool open_only_one_table(THD* thd, TABLE_LIST* table,
                                 bool is_view_operator_func)
 {
   LEX *lex= thd->lex;
-  SELECT_LEX *select= &lex->select_lex;
+  SELECT_LEX *select= lex->first_select_lex();
   TABLE_LIST *save_next_global, *save_next_local;
   bool open_error;
   save_next_global= table->next_global;
@@ -337,19 +356,20 @@ static bool open_only_one_table(THD* thd, TABLE_LIST* table,
     to differentiate from ALTER TABLE...CHECK PARTITION on which view is not
     allowed.
   */
-  if (lex->alter_info.flags & Alter_info::ALTER_ADMIN_PARTITION ||
+  if (lex->alter_info.partition_flags & ALTER_PARTITION_ADMIN ||
       !is_view_operator_func)
   {
-    table->required_type=FRMTYPE_TABLE;
-    DBUG_ASSERT(!lex->only_view);
+    table->required_type= TABLE_TYPE_NORMAL;
+    DBUG_ASSERT(lex->table_type != TABLE_TYPE_VIEW);
   }
-  else if (lex->only_view)
+  else if (lex->table_type == TABLE_TYPE_VIEW)
   {
-    table->required_type= FRMTYPE_VIEW;
+    table->required_type= lex->table_type;
   }
-  else if (!lex->only_view && lex->sql_command == SQLCOM_REPAIR)
+  else if ((lex->table_type != TABLE_TYPE_VIEW) &&
+           lex->sql_command == SQLCOM_REPAIR)
   {
-    table->required_type= FRMTYPE_TABLE;
+    table->required_type= TABLE_TYPE_NORMAL;
   }
 
   if (lex->sql_command == SQLCOM_CHECK ||
@@ -391,6 +411,7 @@ static bool open_only_one_table(THD* thd, TABLE_LIST* table,
     open_error= (thd->open_temporary_tables(table) ||
                  open_and_lock_tables(thd, table, TRUE, 0));
   }
+
 #ifndef DBUG_OFF
 dbug_err:
 #endif
@@ -415,6 +436,50 @@ dbug_err:
   return open_error;
 }
 
+#ifdef WITH_WSREP
+  /*
+     OPTIMIZE, REPAIR and ALTER may take MDL locks not only for the affected table, but
+     also for the table referenced by foreign key constraint.
+     This wsrep_toi_replication() function handles TOI replication for OPTIMIZE and REPAIR
+     so that certification keys for potential FK parent tables are also appended in the
+     write set.
+     ALTER TABLE case is handled elsewhere.
+  */
+static bool wsrep_toi_replication(THD *thd, TABLE_LIST *tables)
+{
+  if (!WSREP(thd) || !WSREP_CLIENT(thd)) return false;
+
+  LEX *lex= thd->lex;
+  /* only handle OPTIMIZE and REPAIR here */
+  switch (lex->sql_command)
+  {
+  case SQLCOM_OPTIMIZE:
+  case SQLCOM_REPAIR:
+    break;
+  default:
+    return false;
+  }
+
+  close_thread_tables(thd);
+  wsrep::key_array keys;
+
+  wsrep_append_fk_parent_table(thd, tables, &keys);
+
+  /* now TOI replication, with no locks held */
+  if (keys.empty())
+  {
+    WSREP_TO_ISOLATION_BEGIN_WRTCHK(NULL, NULL, tables);
+  } else {
+    WSREP_TO_ISOLATION_BEGIN_FK_TABLES(NULL, NULL, tables, &keys) {
+      return true;
+    }
+  }
+  return false;
+
+ wsrep_error_label:
+  return true;
+}
+#endif /* WITH_WSREP */
 
 /*
   RETURN VALUES
@@ -434,8 +499,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
                               int (handler::*operator_func)(THD *,
                                                             HA_CHECK_OPT *),
                               int (view_operator_func)(THD *, TABLE_LIST*,
-                                                       HA_CHECK_OPT *),
-                              bool is_cmd_replicated)
+                                                       HA_CHECK_OPT *))
 {
   TABLE_LIST *table;
   List<Item> field_list;
@@ -446,8 +510,6 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
   int compl_result_code;
   bool need_repair_or_alter= 0;
   wait_for_commit* suspended_wfc;
-  bool is_table_modified= false;
-
   DBUG_ENTER("mysql_admin_table");
   DBUG_PRINT("enter", ("extra_open_options: %u", extra_open_options));
 
@@ -486,20 +548,25 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
   close_thread_tables(thd);
   for (table= tables; table; table= table->next_local)
     table->table= NULL;
+#ifdef WITH_WSREP
+  if (wsrep_toi_replication(thd, tables))
+  {
+    WSREP_INFO("wsrep TOI replication of has failed, skipping OPTIMIZE");
+    goto err;
+  }
+#endif /* WITH_WSREP */
 
   for (table= tables; table; table= table->next_local)
   {
     char table_name[SAFE_NAME_LEN*2+2];
-    char *db= table->db;
+    const char *db= table->db.str;
     bool fatal_error=0;
     bool open_error;
     bool collect_eis=  FALSE;
     bool open_for_modify= org_open_for_modify;
 
-    DBUG_PRINT("admin", ("table: '%s'.'%s'", table->db, table->table_name));
-    DEBUG_SYNC(thd, "admin_command_kill_before_modify");
-
-    strxmov(table_name, db, ".", table->table_name, NullS);
+    DBUG_PRINT("admin", ("table: '%s'.'%s'", db, table->table_name.str));
+    strxmov(table_name, db, ".", table->table_name.str, NullS);
     thd->open_options|= extra_open_options;
     table->lock_type= lock_type;
     /*
@@ -513,14 +580,6 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
                                 : lock_type >= TL_WRITE_ALLOW_WRITE
                                 ? MDL_SHARED_WRITE : MDL_SHARED_READ);
 
-    if (thd->check_killed())
-    {
-      open_error= false;
-      fatal_error= true;
-      result_code= HA_ADMIN_FAILED;
-      goto send_result;
-    }
-
     /* open only one table from local list of command */
     while (1)
     {
@@ -533,7 +592,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
         If open_and_lock_tables() failed, close_thread_tables() will close
         the table and table->table can therefore be invalid.
       */
-      if (open_error)
+      if (unlikely(open_error))
         table->table= NULL;
 
       /*
@@ -541,7 +600,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
         so any errors opening the table are logical errors.
         In these cases it does not make sense to try to repair.
       */
-      if (open_error && thd->locked_tables_mode)
+      if (unlikely(open_error) && thd->locked_tables_mode)
       {
         result_code= HA_ADMIN_FAILED;
         goto send_result;
@@ -556,7 +615,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
       close_thread_tables(thd);
       table->table= NULL;
       thd->release_transactional_locks();
-      table->mdl_request.init(MDL_key::TABLE, table->db, table->table_name,
+      table->mdl_request.init(MDL_key::TABLE, table->db.str, table->table_name.str,
                               MDL_SHARED_NO_READ_WRITE, MDL_TRANSACTION);
     }
 
@@ -570,7 +629,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
         */
         Alter_info *alter_info= &lex->alter_info;
 
-        if (alter_info->flags & Alter_info::ALTER_ADMIN_PARTITION)
+        if (alter_info->partition_flags & ALTER_PARTITION_ADMIN)
         {
           if (!table->table->part_info)
           {
@@ -601,6 +660,12 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
 #endif
     DBUG_PRINT("admin", ("table: %p", table->table));
 
+    if (table->schema_table)
+    {
+      result_code= HA_ADMIN_NOT_IMPLEMENTED;
+      goto send_result;
+    }
+
     if (prepare_func)
     {
       DBUG_PRINT("admin", ("calling prepare_func"));
@@ -659,12 +724,6 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
       goto send_result;
     }
 
-    if (table->schema_table)
-    {
-      result_code= HA_ADMIN_NOT_IMPLEMENTED;
-      goto send_result;
-    }
-
     if ((table->table->db_stat & HA_READ_ONLY) && open_for_modify)
     {
       /* purecov: begin inspected */
@@ -779,7 +838,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
       }
 
       /*
-        The check for Alter_info::ALTER_ADMIN_PARTITION implements this logic:
+        The check for ALTER_PARTITION_ADMIN implements this logic:
         do not collect EITS STATS for this syntax:
           ALTER TABLE ... ANALYZE PARTITION p
         EITS statistics is global (not per-partition). Collecting global stats
@@ -788,8 +847,8 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
       */
       collect_eis=
         (table->table->s->table_category == TABLE_CATEGORY_USER &&
-        !(lex->alter_info.flags & Alter_info::ALTER_ADMIN_PARTITION) &&
-         (get_use_stat_tables_mode(thd) > NEVER ||
+        !(lex->alter_info.flags & ALTER_PARTITION_ADMIN) &&
+         (check_eits_collection_allowed(thd) ||
           lex->with_persistent_for_clause));
     }
 
@@ -808,7 +867,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
         Here we close and reopen table in read mode because operation of
         collecting statistics is long and it will be better do not block
         the table completely.
-        InnoDB/XtraDB will allow read/write and MyISAM read/insert.
+        InnoDB will allow read/write and MyISAM read/insert.
       */
       trans_commit_stmt(thd);
       trans_commit(thd);
@@ -816,7 +875,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
       close_thread_tables(thd);
       table->table= NULL;
       thd->release_transactional_locks();
-      table->mdl_request.init(MDL_key::TABLE, table->db, table->table_name,
+      table->mdl_request.init(MDL_key::TABLE, table->db.str, table->table_name.str,
                               MDL_SHARED_NO_READ_WRITE, MDL_TRANSACTION);
       table->mdl_request.set_type(MDL_SHARED_READ);
 
@@ -826,7 +885,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
                                       repair_table_use_frm, FALSE);
       thd->open_options&= ~extra_open_options;
 
-      if (!open_error)
+      if (unlikely(!open_error))
       {
         TABLE *tab= table->table;
         Field **field_ptr= tab->field;
@@ -843,7 +902,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
               push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
                                   ER_NO_EIS_FOR_FIELD,
                                   ER_THD(thd, ER_NO_EIS_FOR_FIELD),
-                                  (*field_ptr)->field_name);
+                                  (*field_ptr)->field_name.str);
           }
         }
         else
@@ -1021,7 +1080,7 @@ send_result_message:
       Alter_info *alter_info= &lex->alter_info;
 
       protocol->store(STRING_WITH_LEN("note"), system_charset_info);
-      if (alter_info->flags & Alter_info::ALTER_ADMIN_PARTITION)
+      if (alter_info->partition_flags & ALTER_PARTITION_ADMIN)
       {
         protocol->store(STRING_WITH_LEN(
         "Table does not support optimize on partitions. All partitions "
@@ -1060,15 +1119,15 @@ send_result_message:
         if (!thd->open_temporary_tables(table) &&
             (table->table= open_ltable(thd, table, lock_type, 0)))
         {
-          uint save_flags;
+          ulonglong save_flags;
           /* Store the original value of alter_info->flags */
           save_flags= alter_info->flags;
 
           /*
-           Reset the ALTER_ADMIN_PARTITION bit in alter_info->flags
+           Reset the ALTER_PARTITION_ADMIN bit in alter_info->flags
            to force analyze on all partitions.
           */
-          alter_info->flags &= ~(Alter_info::ALTER_ADMIN_PARTITION);
+          alter_info->partition_flags &= ~(ALTER_PARTITION_ADMIN);
           result_code= table->table->file->ha_analyze(thd, check_opt);
           if (result_code == HA_ADMIN_ALREADY_DONE)
             result_code= HA_ADMIN_OK;
@@ -1137,11 +1196,11 @@ send_result_message:
       if (what_to_upgrade)
         length= my_snprintf(buf, sizeof(buf),
                             ER_THD(thd, ER_TABLE_NEEDS_UPGRADE),
-                            what_to_upgrade, table->table_name);
+                            what_to_upgrade, table->table_name.str);
       else
         length= my_snprintf(buf, sizeof(buf),
                             ER_THD(thd, ER_TABLE_NEEDS_REBUILD),
-                            table->table_name);
+                            table->table_name.str);
       protocol->store(buf, length, system_charset_info);
       fatal_error=1;
       break;
@@ -1159,13 +1218,6 @@ send_result_message:
         break;
       }
     }
-    /*
-      Admin commands acquire table locks and these locks are not detected by
-      parallel replication deadlock detection-and-handling mechanism. Hence
-      they must be marked as DDL so that they are not scheduled in parallel
-      with conflicting DMLs resulting in deadlock.
-    */
-    thd->transaction.stmt.mark_executed_table_admin_cmd();
     if (table->table && !table->view)
     {
       if (table->table->s->tmp_table)
@@ -1180,7 +1232,7 @@ send_result_message:
       else if (open_for_modify || fatal_error)
       {
         tdc_remove_table(thd, TDC_RT_REMOVE_UNUSED,
-                         table->db, table->table_name, FALSE);
+                         table->db.str, table->table_name.str, FALSE);
         /*
           May be something modified. Consequently, we have to
           invalidate the query cache.
@@ -1201,10 +1253,10 @@ send_result_message:
     }
     else
     {
-      if (trans_commit_stmt(thd))
+      if (trans_commit_stmt(thd) ||
+          (stmt_causes_implicit_commit(thd, CF_IMPLICIT_COMMIT_END) &&
+           trans_commit_implicit(thd)))
         goto err;
-      if (!is_table_modified)
-        is_table_modified= true;
     }
     close_thread_tables(thd);
     thd->release_transactional_locks();
@@ -1227,12 +1279,6 @@ send_result_message:
 
     if (protocol->write())
       goto err;
-    DEBUG_SYNC(thd, "admin_command_kill_after_modify");
-  }
-  if (is_table_modified && is_cmd_replicated && !thd->lex->no_write_to_binlog)
-  {
-    if (write_bin_log(thd, TRUE, thd->query(), thd->query_length()))
-      goto err;
   }
 
   my_eof(thd);
@@ -1271,7 +1317,7 @@ err:
 */
 
 bool mysql_assign_to_keycache(THD* thd, TABLE_LIST* tables,
-			     LEX_STRING *key_cache_name)
+			     const LEX_CSTRING *key_cache_name)
 {
   HA_CHECK_OPT check_opt;
   KEY_CACHE *key_cache;
@@ -1295,7 +1341,7 @@ bool mysql_assign_to_keycache(THD* thd, TABLE_LIST* tables,
   check_opt.key_cache= key_cache;
   DBUG_RETURN(mysql_admin_table(thd, tables, &check_opt,
 				"assign_to_keycache", TL_READ_NO_INSERT, 0, 0,
-				0, 0, &handler::assign_to_keycache, 0, false));
+				0, 0, &handler::assign_to_keycache, 0));
 }
 
 
@@ -1322,14 +1368,14 @@ bool mysql_preload_keys(THD* thd, TABLE_LIST* tables)
   */
   DBUG_RETURN(mysql_admin_table(thd, tables, 0,
 				"preload_keys", TL_READ_NO_INSERT, 0, 0, 0, 0,
-				&handler::preload_keys, 0, false));
+				&handler::preload_keys, 0));
 }
 
 
 bool Sql_cmd_analyze_table::execute(THD *thd)
 {
   LEX *m_lex= thd->lex;
-  TABLE_LIST *first_table= m_lex->select_lex.table_list.first;
+  TABLE_LIST *first_table= m_lex->first_select_lex()->table_list.first;
   bool res= TRUE;
   thr_lock_type lock_type = TL_READ_NO_INSERT;
   DBUG_ENTER("Sql_cmd_analyze_table::execute");
@@ -1337,15 +1383,28 @@ bool Sql_cmd_analyze_table::execute(THD *thd)
   if (check_table_access(thd, SELECT_ACL | INSERT_ACL, first_table,
                          FALSE, UINT_MAX, FALSE))
     goto error;
+  if (thd->has_read_only_protection())
+    goto error;
+
   WSREP_TO_ISOLATION_BEGIN_WRTCHK(NULL, NULL, first_table);
   res= mysql_admin_table(thd, first_table, &m_lex->check_opt,
                          "analyze", lock_type, 1, 0, 0, 0,
-                         &handler::ha_analyze, 0, true);
-  m_lex->select_lex.table_list.first= first_table;
+                         &handler::ha_analyze, 0);
+  /* ! we write after unlocking the table */
+  if (!res && !m_lex->no_write_to_binlog && (!opt_readonly || thd->slave_thread))
+  {
+    /*
+      Presumably, ANALYZE and binlog writing doesn't require synchronization
+    */
+    res= write_bin_log(thd, TRUE, thd->query(), thd->query_length());
+  }
+  m_lex->first_select_lex()->table_list.first= first_table;
   m_lex->query_tables= first_table;
 
+#ifdef WITH_WSREP
+ wsrep_error_label:
+#endif /* WITH_WSREP */
 error:
-WSREP_ERROR_LABEL:
   DBUG_RETURN(res);
 }
 
@@ -1353,7 +1412,7 @@ WSREP_ERROR_LABEL:
 bool Sql_cmd_check_table::execute(THD *thd)
 {
   LEX *m_lex= thd->lex;
-  TABLE_LIST *first_table= m_lex->select_lex.table_list.first;
+  TABLE_LIST *first_table= m_lex->first_select_lex()->table_list.first;
   thr_lock_type lock_type = TL_READ_NO_INSERT;
   bool res= TRUE;
   DBUG_ENTER("Sql_cmd_check_table::execute");
@@ -1361,11 +1420,12 @@ bool Sql_cmd_check_table::execute(THD *thd)
   if (check_table_access(thd, SELECT_ACL, first_table,
                          TRUE, UINT_MAX, FALSE))
     goto error; /* purecov: inspected */
+
   res= mysql_admin_table(thd, first_table, &m_lex->check_opt, "check",
                          lock_type, 0, 0, HA_OPEN_FOR_REPAIR, 0,
-                         &handler::ha_check, &view_check, false);
+                         &handler::ha_check, &view_check);
 
-  m_lex->select_lex.table_list.first= first_table;
+  m_lex->first_select_lex()->table_list.first= first_table;
   m_lex->query_tables= first_table;
 
 error:
@@ -1376,24 +1436,31 @@ error:
 bool Sql_cmd_optimize_table::execute(THD *thd)
 {
   LEX *m_lex= thd->lex;
-  TABLE_LIST *first_table= m_lex->select_lex.table_list.first;
+  TABLE_LIST *first_table= m_lex->first_select_lex()->table_list.first;
   bool res= TRUE;
   DBUG_ENTER("Sql_cmd_optimize_table::execute");
 
   if (check_table_access(thd, SELECT_ACL | INSERT_ACL, first_table,
                          FALSE, UINT_MAX, FALSE))
     goto error; /* purecov: inspected */
-  WSREP_TO_ISOLATION_BEGIN_WRTCHK(NULL, NULL, first_table);
+
   res= (specialflag & SPECIAL_NO_NEW_FUNC) ?
     mysql_recreate_table(thd, first_table, true) :
     mysql_admin_table(thd, first_table, &m_lex->check_opt,
                       "optimize", TL_WRITE, 1, 0, 0, 0,
-                      &handler::ha_optimize, 0, true);
-  m_lex->select_lex.table_list.first= first_table;
+                      &handler::ha_optimize, 0);
+  /* ! we write after unlocking the table */
+  if (!res && !m_lex->no_write_to_binlog && (!opt_readonly || thd->slave_thread))
+  {
+    /*
+      Presumably, OPTIMIZE and binlog writing doesn't require synchronization
+    */
+    res= write_bin_log(thd, TRUE, thd->query(), thd->query_length());
+  }
+  m_lex->first_select_lex()->table_list.first= first_table;
   m_lex->query_tables= first_table;
 
 error:
-WSREP_ERROR_LABEL:
   DBUG_RETURN(res);
 }
 
@@ -1401,24 +1468,30 @@ WSREP_ERROR_LABEL:
 bool Sql_cmd_repair_table::execute(THD *thd)
 {
   LEX *m_lex= thd->lex;
-  TABLE_LIST *first_table= m_lex->select_lex.table_list.first;
+  TABLE_LIST *first_table= m_lex->first_select_lex()->table_list.first;
   bool res= TRUE;
   DBUG_ENTER("Sql_cmd_repair_table::execute");
 
   if (check_table_access(thd, SELECT_ACL | INSERT_ACL, first_table,
                          FALSE, UINT_MAX, FALSE))
     goto error; /* purecov: inspected */
-  WSREP_TO_ISOLATION_BEGIN_WRTCHK(NULL, NULL, first_table);
   res= mysql_admin_table(thd, first_table, &m_lex->check_opt, "repair",
                          TL_WRITE, 1,
                          MY_TEST(m_lex->check_opt.sql_flags & TT_USEFRM),
                          HA_OPEN_FOR_REPAIR, &prepare_for_repair,
-                         &handler::ha_repair, &view_repair, true);
+                         &handler::ha_repair, &view_repair);
 
-  m_lex->select_lex.table_list.first= first_table;
+  /* ! we write after unlocking the table */
+  if (!res && !m_lex->no_write_to_binlog && (!opt_readonly || thd->slave_thread))
+  {
+    /*
+      Presumably, REPAIR and binlog writing doesn't require synchronization
+    */
+    res= write_bin_log(thd, TRUE, thd->query(), thd->query_length());
+  }
+  m_lex->first_select_lex()->table_list.first= first_table;
   m_lex->query_tables= first_table;
 
 error:
-WSREP_ERROR_LABEL:
   DBUG_RETURN(res);
 }
