@@ -181,7 +181,7 @@ bool get_type_attributes_for_tvc(THD *thd,
     Item *item;
     for (uint holder_pos= 0 ; (item= it++); holder_pos++)
     {
-      DBUG_ASSERT(item->fixed);
+      DBUG_ASSERT(item->is_fixed());
       holders[holder_pos].add_argument(item);
     }
   }
@@ -257,7 +257,6 @@ bool table_value_constr::prepare(THD *thd, SELECT_LEX *sl,
                       Item_type_holder(thd, item, holders[pos].type_handler(),
                                        &holders[pos]/*Type_all_attributes*/,
                                        holders[pos].get_maybe_null());
-    new_holder->fix_fields(thd, 0);
     sl->item_list.push_back(new_holder);
   }
   if (arena)
@@ -281,7 +280,7 @@ bool table_value_constr::prepare(THD *thd, SELECT_LEX *sl,
   for (; order; order=order->next)
   {
     Item *order_item= *order->item;
-    if (order_item->type() == Item::INT_ITEM && order_item->basic_const_item())
+    if (order_item->is_order_clause_position())
     {
       uint count= 0;
       if (order->counter_used)
@@ -333,7 +332,7 @@ int table_value_constr::save_explain_data_intern(THD *thd,
 
   explain->select_id= select_lex->select_number;
   explain->select_type= select_lex->type;
-  explain->linkage= select_lex->linkage;
+  explain->linkage= select_lex->get_linkage();
   explain->using_temporary= false;
   explain->using_filesort= false;
   /* Setting explain->message means that all other members are invalid */
@@ -342,13 +341,6 @@ int table_value_constr::save_explain_data_intern(THD *thd,
   if (select_lex->master_unit()->derived)
     explain->connection_type= Explain_node::EXPLAIN_NODE_DERIVED;
 
-  for (SELECT_LEX_UNIT *unit= select_lex->first_inner_unit();
-       unit;
-       unit= unit->next_unit())
-  {
-    explain->add_child(unit->first_select()->select_number);
-  }
-
   output->add_node(explain);
 
   if (select_lex->is_top_level_node())
@@ -373,14 +365,9 @@ bool table_value_constr::optimize(THD *thd)
       thd->lex->explain && // for "SET" command in SPs.
       (!thd->lex->explain->get_select(select_lex->select_number)))
   {
-    if (save_explain_data_intern(thd, thd->lex->explain))
-      return true;
+    return save_explain_data_intern(thd, thd->lex->explain);
   }
-
-  if (select_lex->optimize_unflattened_subqueries(true))
-    return true;
-
-  return false;
+  return 0;
 }
 
 
@@ -539,10 +526,7 @@ bool Item_func_in::create_value_list_for_tvc(THD *thd,
 
     if (is_list_of_rows)
     {
-      Item_row *row_list= (Item_row *)(args[i]->build_clone(thd));
-
-      if (!row_list)
-        return true;
+      Item_row *row_list= (Item_row *)(args[i]);
 
       for (uint j=0; j < row_list->cols(); j++)
       {
@@ -564,8 +548,7 @@ bool Item_func_in::create_value_list_for_tvc(THD *thd,
         sprintf(col_name, "_col_%i", 1);
         args[i]->set_name(thd, col_name, strlen(col_name), thd->charset());
       }
-      Item *arg_clone= args[i]->build_clone(thd);
-      if (!arg_clone || tvc_value->push_back(arg_clone))
+      if (tvc_value->push_back(args[i]->real_item()))
         return true;
     }
 
@@ -624,8 +607,8 @@ static bool create_tvc_name(THD *thd, st_select_lex *parent_select,
 
 bool table_value_constr::to_be_wrapped_as_with_tail()
 {
-  return select_lex->master_unit()->first_select()->next_select() &&
-         select_lex->order_list.elements && select_lex->explicit_limit;
+  return  select_lex->master_unit()->first_select()->next_select() &&
+          select_lex->order_list.elements && select_lex->explicit_limit;
 }
 
 
@@ -652,69 +635,50 @@ st_select_lex *wrap_tvc(THD *thd, st_select_lex *tvc_sl,
                         st_select_lex *parent_select)
 {
   LEX *lex= thd->lex;
-  select_result *save_result= lex->result;
+  select_result *save_result= thd->lex->result;
   uint8 save_derived_tables= lex->derived_tables;
   thd->lex->result= NULL;
 
   Query_arena backup;
   Query_arena *arena= thd->activate_stmt_arena_if_needed(&backup);
-
+  /*
+    Create SELECT_LEX of the select used in the result of transformation
+  */
+  lex->current_select= tvc_sl;
+  if (mysql_new_select(lex, 0, NULL))
+    goto err;
+  mysql_init_select(lex);
+  /* Create item list as '*' for the subquery SQ */
   Item *item;
   SELECT_LEX *wrapper_sl;
-  SELECT_LEX_UNIT *derived_unit;
-
-  /*
-    Create SELECT_LEX wrapper_sl of the select used in the result
-    of the transformation
-  */
-  if (!(wrapper_sl= new (thd->mem_root) SELECT_LEX()))
-    goto err;
-  wrapper_sl->select_number= ++thd->lex->stmt_lex->current_select_number;
-  wrapper_sl->parent_lex= lex; /* Used in init_query. */
-  wrapper_sl->init_query();
-  wrapper_sl->init_select();
-
-  wrapper_sl->nest_level= tvc_sl->nest_level;
-  wrapper_sl->parsing_place= tvc_sl->parsing_place;
-  wrapper_sl->linkage= tvc_sl->linkage;
-  wrapper_sl->exclude_from_table_unique_test=
-                                 tvc_sl->exclude_from_table_unique_test;
-
-  lex->current_select= wrapper_sl;
+  wrapper_sl= lex->current_select;
+  wrapper_sl->set_linkage(tvc_sl->get_linkage());
+  wrapper_sl->parsing_place= SELECT_LIST;
   item= new (thd->mem_root) Item_field(thd, &wrapper_sl->context,
                                        NULL, NULL, &star_clex_str);
   if (item == NULL || add_item_to_list(thd, item))
     goto err;
   (wrapper_sl->with_wild)++;
-
-  /* Include the newly created select into the global list of selects */
-  wrapper_sl->include_global((st_select_lex_node**)&lex->all_selects_list);
-
-  /* Substitute select node used of TVC for the newly created select */
-  tvc_sl->substitute_in_tree(wrapper_sl);
-
+  
+  /* Exclude SELECT with TVC */
+  tvc_sl->exclude();
   /*
-    Create a unit for the substituted select used for TVC and attach it
-    to the the wrapper select wrapper_sl as the only unit. The created
-    unit is the unit for the derived table tvc_x of the transformation.
+    Create derived table DT that will wrap TVC in the result of transformation
   */
-  if (!(derived_unit= new (thd->mem_root) SELECT_LEX_UNIT()))
+  SELECT_LEX *tvc_select; // select for tvc
+  SELECT_LEX_UNIT *derived_unit; // unit for tvc_select
+  if (mysql_new_select(lex, 1, tvc_sl))
     goto err;
-  derived_unit->init_query();
-  derived_unit->thd= thd;
-  derived_unit->include_down(wrapper_sl);
+  tvc_select= lex->current_select;
+  derived_unit= tvc_select->master_unit();
+  tvc_select->set_linkage(DERIVED_TABLE_TYPE);
+
+  lex->current_select= wrapper_sl;
 
   /*
-    Attach the select used of TVC as the only slave to the unit for
-    the derived table tvc_x of the transformation
-  */
-  derived_unit->attach_single(tvc_sl);
-  tvc_sl->linkage= DERIVED_TABLE_TYPE;
-
-  /*
-    Generate the name of the derived table created for TVC and
-    add it to the FROM list of the wrapping select
-  */
+    Create the name of the wrapping derived table and
+    add it to the FROM list of the wrapper
+   */
   Table_ident *ti;
   LEX_CSTRING alias;
   TABLE_LIST *derived_tab;
@@ -733,15 +697,19 @@ st_select_lex *wrap_tvc(THD *thd, st_select_lex *tvc_sl,
   wrapper_sl->table_list.first->derived_type= DTYPE_TABLE | DTYPE_MATERIALIZE;
   lex->derived_tables|= DERIVED_SUBQUERY;
 
+  wrapper_sl->where= 0;
+  wrapper_sl->set_braces(false);
+  derived_unit->set_with_clause(0);
+
   if (arena)
     thd->restore_active_arena(arena, &backup);
-  lex->result= save_result;
+  thd->lex->result= save_result;
   return wrapper_sl;
 
 err:
   if (arena)
     thd->restore_active_arena(arena, &backup);
-  lex->result= save_result;
+  thd->lex->result= save_result;
   lex->derived_tables= save_derived_tables;
   return 0;
 }
@@ -810,12 +778,11 @@ st_select_lex *wrap_tvc_with_tail(THD *thd, st_select_lex *tvc_sl)
     SELECT * FROM (VALUES (v1), ... (vn)) tvc_x
     and replaces the subselect with the result of the transformation.
 
-  @retval wrapping select if successful
-          0  otherwise
+  @retval false if successfull
+          true  otherwise
 */
 
-st_select_lex *
-Item_subselect::wrap_tvc_into_select(THD *thd, st_select_lex *tvc_sl)
+bool Item_subselect::wrap_tvc_into_select(THD *thd, st_select_lex *tvc_sl)
 {
   LEX *lex= thd->lex;
   /* SELECT_LEX object where the transformation is performed */
@@ -825,9 +792,14 @@ Item_subselect::wrap_tvc_into_select(THD *thd, st_select_lex *tvc_sl)
   {
     if (engine->engine_type() == subselect_engine::SINGLE_SELECT_ENGINE)
       ((subselect_single_select_engine *) engine)->change_select(wrapper_sl);
+    lex->current_select= wrapper_sl;
+    return false;
+  }
+  else
+  {
+    lex->current_select= parent_select;
+    return true;
   }
-  lex->current_select= parent_select;
-  return wrapper_sl;
 }
 
 
@@ -900,6 +872,8 @@ Item *Item_func_in::in_predicate_to_in_subs_transformer(THD *thd,
   if (!transform_into_subq)
     return this;
   
+  transform_into_subq= false;
+
   List<List_item> values;
 
   LEX *lex= thd->lex;
@@ -954,7 +928,7 @@ Item *Item_func_in::in_predicate_to_in_subs_transformer(THD *thd,
   mysql_init_select(lex);
   tvc_select= lex->current_select;
   derived_unit= tvc_select->master_unit();
-  tvc_select->linkage= DERIVED_TABLE_TYPE;
+  tvc_select->set_linkage(DERIVED_TABLE_TYPE);
 
   /* Create TVC used in the transformation */
   if (create_value_list_for_tvc(thd, &values))
@@ -1056,38 +1030,15 @@ uint32 Item_func_in::max_length_of_left_expr()
 
 bool Item_func_in::to_be_transformed_into_in_subq(THD *thd)
 {
-  bool is_row_list= args[1]->type() == Item::ROW_ITEM;
   uint values_count= arg_count-1;
 
-  if (is_row_list)
+  if (args[1]->type() == Item::ROW_ITEM)
     values_count*= ((Item_row *)(args[1]))->cols();
 
   if (thd->variables.in_subquery_conversion_threshold == 0 ||
       thd->variables.in_subquery_conversion_threshold > values_count)
     return false;
 
-  if (!(thd->lex->context_analysis_only & CONTEXT_ANALYSIS_ONLY_PREPARE))
-    return true;
-
-  /* Occurence of '?' in IN list is checked only for PREPARE <stmt> commands */
-  for (uint i=1; i < arg_count; i++)
-  {
-    if (!is_row_list)
-    {
-      if (args[i]->type() == Item::PARAM_ITEM)
-        return false;
-    }
-    else
-    {
-      Item_row *row_list= (Item_row *)(args[i]);
-      for (uint j=0; j < row_list->cols(); j++)
-      {
-        if (row_list->element_index(j)->type() == Item::PARAM_ITEM)
-          return false;
-      }
-    }
-  }
-
   return true;
 }
 
