@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 2016, 2018, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2017, 2021, MariaDB Corporation.
+Copyright (c) 2017, 2020, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -145,7 +145,7 @@ rtr_pcur_getnext_from_path(
 						| MTR_MEMO_X_LOCK));
 	}
 
-	const page_size_t	page_size(index->table->space->flags);
+	const ulint zip_size = index->table->space->zip_size();
 
 	/* Pop each node/page to be searched from "path" structure
 	and do a search on it. Please note, any pages that are in
@@ -269,7 +269,7 @@ rtr_pcur_getnext_from_path(
 
 		block = buf_page_get_gen(
 			page_id_t(index->table->space_id,
-				  next_rec.page_no), page_size,
+				  next_rec.page_no), zip_size,
 			rw_latch, NULL, BUF_GET, __FILE__, __LINE__, mtr, &err);
 
 		if (block == NULL) {
@@ -424,7 +424,7 @@ rtr_pcur_getnext_from_path(
 						block,
 						page_id_t(index->table->space_id,
 							  block->page.id.page_no()),
-						page_size, BTR_MODIFY_TREE,
+						zip_size, BTR_MODIFY_TREE,
 						btr_cur, mtr);
 				}
 
@@ -530,7 +530,8 @@ rtr_compare_cursor_rec(
 
 	rec = btr_cur_get_rec(cursor);
 
-	offsets = rec_get_offsets(rec, index, NULL, 0, ULINT_UNDEFINED, heap);
+	offsets = rec_get_offsets(
+		rec, index, NULL, false, ULINT_UNDEFINED, heap);
 
 	return(btr_node_ptr_get_child_page_no(rec, offsets) == page_no);
 }
@@ -835,8 +836,7 @@ rtr_page_get_father_node_ptr(
 	user_rec = btr_cur_get_rec(cursor);
 	ut_a(page_rec_is_user_rec(user_rec));
 
-	offsets = rec_get_offsets(user_rec, index, offsets,
-				  level ? 0 : index->n_fields,
+	offsets = rec_get_offsets(user_rec, index, offsets, !level,
 				  ULINT_UNDEFINED, &heap);
 	rtr_get_mbr_from_rec(user_rec, offsets, &mbr);
 
@@ -853,7 +853,7 @@ rtr_page_get_father_node_ptr(
 	node_ptr = btr_cur_get_rec(cursor);
 	ut_ad(!page_rec_is_comp(node_ptr)
 	      || rec_get_status(node_ptr) == REC_STATUS_NODE_PTR);
-	offsets = rec_get_offsets(node_ptr, index, offsets, 0,
+	offsets = rec_get_offsets(node_ptr, index, offsets, false,
 				  ULINT_UNDEFINED, &heap);
 
 	ulint	child_page = btr_node_ptr_get_child_page_no(node_ptr, offsets);
@@ -871,14 +871,13 @@ rtr_page_get_father_node_ptr(
 		print_rec = page_rec_get_next(
 			page_get_infimum_rec(page_align(user_rec)));
 		offsets = rec_get_offsets(print_rec, index, offsets,
-					  page_rec_is_leaf(user_rec)
-					  ? index->n_fields : 0,
+					  page_rec_is_leaf(user_rec),
 					  ULINT_UNDEFINED, &heap);
 		error << "; child ";
 		rec_print(error.m_oss, print_rec,
 			  rec_get_info_bits(print_rec, rec_offs_comp(offsets)),
 			  offsets);
-		offsets = rec_get_offsets(node_ptr, index, offsets, 0,
+		offsets = rec_get_offsets(node_ptr, index, offsets, false,
 					  ULINT_UNDEFINED, &heap);
 		error << "; parent ";
 		rec_print(error.m_oss, print_rec,
@@ -971,7 +970,7 @@ rtr_create_rtr_info(
 		     &rtr_info->rtr_path_mutex);
 
 	mutex_enter(&index->rtr_track->rtr_active_mutex);
-	index->rtr_track->rtr_active->push_back(rtr_info);
+	index->rtr_track->rtr_active.push_front(rtr_info);
 	mutex_exit(&index->rtr_track->rtr_active_mutex);
 	return(rtr_info);
 }
@@ -1044,7 +1043,7 @@ rtr_init_rtr_info(
 	rtr_info->index = index;
 
 	mutex_enter(&index->rtr_track->rtr_active_mutex);
-	index->rtr_track->rtr_active->push_back(rtr_info);
+	index->rtr_track->rtr_active.push_front(rtr_info);
 	mutex_exit(&index->rtr_track->rtr_active_mutex);
 }
 
@@ -1096,7 +1095,7 @@ rtr_clean_rtr_info(
 	}
 
 	if (index) {
-		index->rtr_track->rtr_active->remove(rtr_info);
+		index->rtr_track->rtr_active.remove(rtr_info);
 		mutex_exit(&index->rtr_track->rtr_active_mutex);
 	}
 
@@ -1201,36 +1200,22 @@ rtr_check_discard_page(
 				the root page */
 	buf_block_t*	block)	/*!< in: block of page to be discarded */
 {
-	ulint			pageno = block->page.id.page_no();
-	rtr_info_t*		rtr_info;
-	rtr_info_active::iterator	it;
+	const ulint pageno = block->page.id.page_no();
 
 	mutex_enter(&index->rtr_track->rtr_active_mutex);
 
-	for (it = index->rtr_track->rtr_active->begin();
-	     it != index->rtr_track->rtr_active->end(); ++it) {
-		rtr_info = *it;
-		rtr_node_path_t::iterator	rit;
-		bool	found = false;
-
+	for (const auto& rtr_info : index->rtr_track->rtr_active) {
 		if (cursor && rtr_info == cursor->rtr_info) {
 			continue;
 		}
 
 		mutex_enter(&rtr_info->rtr_path_mutex);
-		for (rit = rtr_info->path->begin();
-		     rit != rtr_info->path->end(); ++rit) {
-			node_visit_t	node = *rit;
-
+		for (const node_visit_t& node : *rtr_info->path) {
 			if (node.page_no == pageno) {
-				found = true;
+				rtr_rebuild_path(rtr_info, pageno);
 				break;
 			}
 		}
-
-		if (found) {
-			rtr_rebuild_path(rtr_info, pageno);
-		}
 		mutex_exit(&rtr_info->rtr_path_mutex);
 
 		if (rtr_info->matches) {
@@ -1324,12 +1309,10 @@ rtr_cur_restore_position(
 
 			heap = mem_heap_create(256);
 			offsets1 = rec_get_offsets(
-				r_cursor->old_rec, index, NULL,
-				level ? 0 : r_cursor->old_n_fields,
+				r_cursor->old_rec, index, NULL, !level,
 				r_cursor->old_n_fields, &heap);
 			offsets2 = rec_get_offsets(
-				rec, index, NULL,
-				level ? 0 : r_cursor->old_n_fields,
+				rec, index, NULL, !level,
 				r_cursor->old_n_fields, &heap);
 
 			comp = rec_offs_comp(offsets1);
@@ -1360,8 +1343,7 @@ rtr_cur_restore_position(
 	page_cur_t*	page_cursor;
 	node_visit_t*	node = rtr_get_parent_node(btr_cur, level, false);
 	node_seq_t	path_ssn = node->seq_no;
-	const page_size_t	page_size(index->table->space->flags);
-
+	const ulint	zip_size = index->table->space->zip_size();
 	ulint		page_no = node->page_no;
 
 	heap = mem_heap_create(256);
@@ -1377,7 +1359,7 @@ search_again:
 
 	block = buf_page_get_gen(
 		page_id_t(index->table->space_id, page_no),
-		page_size, RW_X_LATCH, NULL,
+		zip_size, RW_X_LATCH, NULL,
 		BUF_GET, __FILE__, __LINE__, mtr, &err);
 
 	ut_ad(block);
@@ -1397,12 +1379,12 @@ search_again:
 
 		rec = btr_pcur_get_rec(r_cursor);
 
-		offsets1 = rec_get_offsets(r_cursor->old_rec, index, NULL,
-					   level ? 0 : r_cursor->old_n_fields,
-					   r_cursor->old_n_fields, &heap);
-		offsets2 = rec_get_offsets(rec, index, NULL,
-					   level ? 0 : r_cursor->old_n_fields,
-					   r_cursor->old_n_fields, &heap);
+		offsets1 = rec_get_offsets(
+			r_cursor->old_rec, index, NULL, !level,
+			r_cursor->old_n_fields, &heap);
+		offsets2 = rec_get_offsets(
+			rec, index, NULL, !level,
+			r_cursor->old_n_fields, &heap);
 
 		comp = rec_offs_comp(offsets1);
 
@@ -1577,14 +1559,13 @@ rtr_copy_buf(
 	matches->block.n_fields = block->n_fields;
 	matches->block.left_side = block->left_side;
 #if defined UNIV_AHI_DEBUG || defined UNIV_DEBUG
-	matches->block.n_pointers = block->n_pointers;
+	matches->block.n_pointers = 0;
 #endif /* UNIV_AHI_DEBUG || UNIV_DEBUG */
 	matches->block.curr_n_fields = block->curr_n_fields;
 	matches->block.curr_left_side = block->curr_left_side;
 	matches->block.index = block->index;
 #endif /* BTR_CUR_HASH_ADAPT */
-	ut_d(matches->block.debug_latch = block->debug_latch);
-
+	ut_d(matches->block.debug_latch = NULL);
 }
 
 /****************************************************************//**
@@ -1692,7 +1673,7 @@ rtr_cur_search_with_match(
 	page = buf_block_get_frame(block);
 
 	const ulint level = btr_page_get_level(page);
-	const ulint n_core = level ? 0 : index->n_fields;
+	const bool is_leaf = !level;
 
 	if (mode == PAGE_CUR_RTREE_LOCATE) {
 		ut_ad(level != 0);
@@ -1714,7 +1695,7 @@ rtr_cur_search_with_match(
 
 		ulint	new_rec_size = rec_get_converted_size(index, tuple, 0);
 
-		offsets = rec_get_offsets(rec, index, offsets, n_core,
+		offsets = rec_get_offsets(rec, index, offsets, is_leaf,
 					  dtuple_get_n_fields_cmp(tuple),
 					  &heap);
 
@@ -1735,10 +1716,10 @@ rtr_cur_search_with_match(
 	}
 
 	while (!page_rec_is_supremum(rec)) {
-		offsets = rec_get_offsets(rec, index, offsets, n_core,
+		offsets = rec_get_offsets(rec, index, offsets, is_leaf,
 					  dtuple_get_n_fields_cmp(tuple),
 					  &heap);
-		if (!n_core) {
+		if (!is_leaf) {
 			switch (mode) {
 			case PAGE_CUR_CONTAIN:
 			case PAGE_CUR_INTERSECT:
@@ -1819,7 +1800,7 @@ rtr_cur_search_with_match(
 			to rtr_info->path for non-leaf nodes, or
 			rtr_info->matches for leaf nodes */
 			if (rtr_info && mode != PAGE_CUR_RTREE_INSERT) {
-				if (!n_core) {
+				if (!is_leaf) {
 					ulint		page_no;
 					node_seq_t	new_seq;
 					bool		is_loc;
@@ -1830,7 +1811,7 @@ rtr_cur_search_with_match(
 						  == PAGE_CUR_RTREE_GET_FATHER);
 
 					offsets = rec_get_offsets(
-						rec, index, offsets, 0,
+						rec, index, offsets, false,
 						ULINT_UNDEFINED, &heap);
 
 					page_no = btr_node_ptr_get_child_page_no(
@@ -1879,8 +1860,7 @@ rtr_cur_search_with_match(
 
 					/* Collect matched records on page */
 					offsets = rec_get_offsets(
-						rec, index, offsets,
-						index->n_fields,
+						rec, index, offsets, true,
 						ULINT_UNDEFINED, &heap);
 					rtr_leaf_push_match_rec(
 						rec, rtr_info, offsets,
@@ -1903,7 +1883,7 @@ rtr_cur_search_with_match(
 
 	/* All records on page are searched */
 	if (page_rec_is_supremum(rec)) {
-		if (!n_core) {
+		if (!is_leaf) {
 			if (!found) {
 				/* No match case, if it is for insertion,
 				then we select the record that result in
@@ -1913,7 +1893,7 @@ rtr_cur_search_with_match(
 					ut_ad(least_inc < DBL_MAX);
 					offsets = rec_get_offsets(
 						best_rec, index, offsets,
-						0, ULINT_UNDEFINED, &heap);
+						false, ULINT_UNDEFINED, &heap);
 					child_no =
 					btr_node_ptr_get_child_page_no(
 						best_rec, offsets);
@@ -1965,11 +1945,11 @@ rtr_cur_search_with_match(
 			/* Verify the record to be positioned is the same
 			as the last record in matched_rec vector */
 			offsets2 = rec_get_offsets(test_rec.r_rec, index,
-						   offsets2, index->n_fields,
+						   offsets2, true,
 						   ULINT_UNDEFINED, &heap);
 
 			offsets = rec_get_offsets(last_match_rec, index,
-						  offsets, index->n_fields,
+						  offsets, true,
 						  ULINT_UNDEFINED, &heap);
 
 			ut_ad(cmp_rec_rec(test_rec.r_rec, last_match_rec,
@@ -1986,8 +1966,9 @@ rtr_cur_search_with_match(
 			ulint	child_no;
 			ut_ad(!last_match_rec && rec);
 
-			offsets = rec_get_offsets(rec, index, offsets, 0,
-						  ULINT_UNDEFINED, &heap);
+			offsets = rec_get_offsets(
+				rec, index, offsets, false,
+				ULINT_UNDEFINED, &heap);
 
 			child_no = btr_node_ptr_get_child_page_no(rec, offsets);
 
@@ -1995,7 +1976,7 @@ rtr_cur_search_with_match(
 				index, rtr_info->parent_path, level, child_no,
 				block, rec, 0);
 
-		} else if (rtr_info && found && !n_core) {
+		} else if (rtr_info && found && !is_leaf) {
 			rec = last_match_rec;
 		}
 
@@ -2005,11 +1986,11 @@ rtr_cur_search_with_match(
 #ifdef UNIV_DEBUG
 	/* Verify that we are positioned at the same child page as pushed in
 	the path stack */
-	if (!n_core && (!page_rec_is_supremum(rec) || found)
+	if (!is_leaf && (!page_rec_is_supremum(rec) || found)
 	    && mode != PAGE_CUR_RTREE_INSERT) {
 		ulint		page_no;
 
-		offsets = rec_get_offsets(rec, index, offsets, 0,
+		offsets = rec_get_offsets(rec, index, offsets, false,
 					  ULINT_UNDEFINED, &heap);
 		page_no = btr_node_ptr_get_child_page_no(rec, offsets);
 
